#define __USE_XOPEN2K
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <assert.h>

#include "socket_wrapper.h"
#include "libgetopt.h"
#include "renju.h"

typedef struct {
    int player; // 0 or 1
    int fd;
}player_args_t;

static int p_player_turn; // -1: game end, 0: player0 turn, 1: player1 turn
static int p_max_sleep_times; // wait p_max_sleep_times * 10ms
static int p_usleep_time; // 10,000 usec
static int p_winner_player;
static pthread_barrier_t p_turn_barrier;

static int put_stone_auto(int player)
{
    int x, y, i;
    unsigned char* data = renju_get_board_data();
    int w = renju_get_board_width();
    int h = renju_get_board_height();
    for(i = y = 0; y < h; y++)
        for(x = 0; x < w; x++, i++)
            if(data[i] == EMPTY_GRID)
            {
                int ret = renju_put_stone(x, y, player);
//                printf("%d %d %d\n", x, y, ret);
                if( (ret == PUT_SUCCESS) || (ret == WINNER) )
                    return ret;
            }
    return END_GAME;
}
static int receive_command(int fd, int default_flag, int player)
{
    char cmd[128];
    int ret = NOP;
    int recv_bytes;
    memset(cmd, 0, sizeof(cmd));
    recv_bytes = read(fd, cmd, sizeof(cmd));
    if( recv_bytes < 0 )
        ;
    else if(!strncmp(cmd, "end", 3))
    {
        ret = END_GAME;
    }
    else if(!strncmp(cmd, "get", 3))
    {
        if(!strncmp(cmd, "get text", 8))
            write(fd, renju_get_board_data_text(), renju_get_board_text_size());
        else if(!strncmp(cmd, "get bin", 7))
            write(fd, renju_get_board_data(), renju_get_board_size());
    }
    else if(!strncmp(cmd, "put", 3))
    {
        int x, y;
        sscanf(cmd, "put %d %d", &x, &y);
        ret = renju_put_stone(x, y, player);
        memset(cmd, 0, sizeof(cmd));
        sprintf(cmd,"%d\n",ret);
        write(fd, cmd, sizeof(cmd));
    }
/*
    else if( recv_bytes > 0 )
    {
        char error_msg[128];
        sprintf(error_msg, "No such command is defined. %s\n", cmd);
        write(fd, error_msg, sizeof(error_msg));
    }
*/
    return ret;
}
#define SEND_KEYWORD(fd,str) write(fd, str, sizeof(str))
        
void* renju_thread(void* arg)
{
    player_args_t* p_args = (player_args_t*)arg;
    int player = p_args->player;
    int fd = p_args->fd;
    int num_of_sleep = 0;
    int default_flag = fcntl(fd, F_GETFL);

    fcntl(fd, F_SETFL, O_NONBLOCK);
    if( player == 0)
    {
        printf("player 0 turn\n");
        SEND_KEYWORD(fd,"your turn\n");
        renju_display_borad(10);
    }
    while(1)
    {
        if( player != p_player_turn )
        {
            pthread_barrier_wait(&p_turn_barrier);
            num_of_sleep = 0;
            printf("player %d turn\n", player);
            SEND_KEYWORD(fd,"your turn\n");
        }
        if( p_player_turn == -1 )
            break;

        int ret = receive_command(fd, default_flag, player);
        if((ret == NOP)&&(num_of_sleep > p_max_sleep_times))
            ret = put_stone_auto(player);
        
        num_of_sleep++;
        if( ret == PUT_SUCCESS )
        {
            renju_display_borad(10);
            p_player_turn ^= 1;
            pthread_barrier_wait(&p_turn_barrier);
        }
        else if( (ret == END_GAME) || (ret == WINNER) )
        {
            if( ret == WINNER )
                p_winner_player = player;
            p_player_turn = -1;
            pthread_barrier_wait(&p_turn_barrier);
            break;
        }
        usleep(p_usleep_time);
    }

    printf("player %d finished\n", player);
    if( p_winner_player == -1 )
        SEND_KEYWORD(fd,"draw game\n");
    else if( p_winner_player == player )
        SEND_KEYWORD(fd,"you win!\n");
    else
        SEND_KEYWORD(fd,"you lose!\n");
    close(p_args->fd);
    return NULL;
}
int main(int argc, char** argv)
{
    getopt_init("Renju socket server", "0.9");
    getopt_setarg("sleep", 1, NULL, 's', "number of 10ms sleep operations during waiting player response.", NULL);
    getopt_setarg("board", 1, NULL, 'b', "board texture image.", NULL);
    getopt_setarg("player0", 1, NULL, '0', "player 0 port number.", NULL);
    getopt_setarg("player1", 1, NULL, '1', "player 1 port number.", NULL);
    getopt_aggregate(argc, argv);

    int player0_port = 10000;
    int player1_port = 10001;
    sockwrap_t sock[2];
    pthread_t player_thread[3];
    player_args_t player_args[2];

    GETOPT_GETARG("player0", player0_port, atoi);
    GETOPT_GETARG("player1", player1_port, atoi);

    sock[0] = init_server_socket(TCP_SOCKET, player0_port);
    sock[1] = init_server_socket(TCP_SOCKET, player1_port);
    player_args[0].fd = accept_socket(&sock[0]);
    player_args[1].fd = accept_socket(&sock[1]);

    char* board_name = getopt_getarg("board");
    renju_initialize_board(15,15,15,board_name);
    p_max_sleep_times = 1000;
    p_usleep_time = 10000;
    p_player_turn = 0;
    p_winner_player = -1;

    GETOPT_GETARG("sleep", p_max_sleep_times, atoi);

    pthread_barrier_init(&p_turn_barrier, NULL, 2);

    player_args[0].player = 0;
    player_args[1].player = 1;
    pthread_create(&player_thread[0], NULL, renju_thread, &player_args[0]);
    pthread_create(&player_thread[1], NULL, renju_thread, &player_args[1]);

    pthread_join(player_thread[0], NULL);
    pthread_join(player_thread[1], NULL);

    renju_save_borad();
    if( p_winner_player == -1 )
        printf("draw game\n");
    else
        printf("winner player %d\n", p_winner_player);
    renju_draw_winpoint(p_winner_player);
//    renju_display_borad(0);
    return 0;
}
